

//#include <stdlib.h>
//#include <iostream.h>
//#include <stdarg.h>
extern "C" {
  int printf(char*, ...) ;
}
extern "C" {
  #include "mpi.h"
}
#include "Seq.h"
#include "Procs.h"
#include "Range.h"
#include "Array.h"
#include "Threads.h"
#include "ProcTree.h"
#include "DataExchange.h"
#include "remap.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RemapState'.                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


RemapState :: RemapState(const int _typeLen, const int _rank,
			 Range** _locRngs, Stride** _locStrs,
			 Range** _remRngs, Stride** _remStrs,
			 ProcessGroup* _remProcGrp,
			 DataExchange* _exchange,
                         Array1<int>** _subscripts) :
	      typeLen(_typeLen), rank(_rank),
	      locRngs(_locRngs), locStrs(_locStrs),
	      remRngs(_remRngs), remStrs(_remStrs),
	      remProcGrp(_remProcGrp), exchange(_exchange),
              subscripts(_subscripts) {

  exts     = new int [rank] ;
  locSteps = new int [rank] ;
  remSteps = new int [rank] ;
}

RemapState :: ~RemapState() {
  delete remSteps ;
  delete locSteps ;
  delete exts ;
}


void RemapState :: remapGetSubrange(char* locData, int locStep,
                          int remOffset, Range* remRng, Stride* remStr,
                          const int d) {

  for(remRng->allBlocksEnum() ; remRng->allBlocksTest() ;
			        remRng->allBlocksNext()) {
    const int iCount = remRng->blockCount() ;

    const int remOffsetElem = remOffset + remStr->offset(remRng) ;
    const int remStep       = remStr->step(remRng) ;

    if(remRng->distributed())
      remProcGrp->deactivate(remRng) ;

    exts     [d] = iCount ;
    remSteps [d] = remStep ;

    remapLoop(locData, remOffsetElem, d + 1) ;

    if(remRng->distributed())
      remProcGrp->reactivate(remRng) ;

    locData += iCount * locStep ;
  }
}

void RemapState :: remapLoop(char* locData, int remOffset, const int d) {

  if(d == rank) {
    exchange->getArray(typeLen, rank, exts,
		       locData, locSteps,
		       remProcGrp->procId(), NULL, remOffset, remSteps) ;
  }
  else {

    Range*  x      = locRngs [d] ;
    Stride* locStr = locStrs [d] ;

    Range*  y      = remRngs [d] ;
    Stride* remStr = remStrs [d] ;

    if(x->distributed())
      apg->deactivate(x) ;

    for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
      char* locDataElem = locData + locStr->offset(x) ;
      const int locStep = locStr->step(x) ;

      locSteps [d] = locStep ;

      Range ySub(x->blockCount(), x->blockBase(), 1, *y) ;

      remapGetSubrange(locDataElem, locStep, remOffset, &ySub, remStr, d) ;
    }

    if(x->distributed())
      apg->reactivate(x) ;
  }
}


void RemapState :: vecGatherLoop(char* locData, int remOffset,
                                 const int d, const int r) {

  if(d == rank)
    exchange->getArray(typeLen, r, exts,
		       locData, locSteps,
		       remProcGrp->procId(), NULL, remOffset, remSteps) ;
  else {

    Range*  x      = locRngs [d] ;
    Stride* locStr = locStrs [d] ;

    Range*  y      = remRngs [d] ;
    Stride* remStr = remStrs [d] ;

    Array1<int>* subscript = subscripts [d] ;

    if(subscript) {

      // Dimension has a vector subscript

      for(x->where() ; (*x) ; (*x)++) {
	char* locDataElem = locData + locStr->offset(x) ;

        Embed ySub((*subscript)(*x), *y) ;

	if(y->distributed())
	  remProcGrp->deactivate(&ySub) ;

	int remOffsetElem = remOffset + remStr->offset(&ySub) ;

	vecGatherLoop(locDataElem, remOffsetElem, d + 1, r) ;

	if(y->distributed())
	  remProcGrp->reactivate(&ySub) ;
      } x->erewh() ;
    }
    else {

      // Dimension is unsubscripted (essentially identical to `remap').

      if(x->distributed())
	apg->deactivate(x) ;

      for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
	char* locDataElem = locData + locStr->offset(x) ;
	const int locStep = locStr->step(x) ;

	locSteps [r] = locStep ;

	Range ySub(x->blockCount(), x->blockBase(), 1, *y) ;

	for(ySub.allBlocksEnum() ; ySub.allBlocksTest() ;
				   ySub.allBlocksNext()) {
	  const int iCount = ySub.blockCount() ;

	  const int remOffsetElem = remOffset + remStr->offset(&ySub) ;
	  const int remStep       = remStr->step(&ySub) ;

	  if(y->distributed())
	    remProcGrp->deactivate(&ySub) ;

	  exts     [r] = iCount ;
	  remSteps [r] = remStep ;

	  vecGatherLoop(locDataElem, remOffsetElem, d + 1, r + 1) ;

	  if(y->distributed())
	    remProcGrp->reactivate(&ySub) ;

	  locDataElem += iCount * locStep ;
	}
      }

      if(x->distributed())
	apg->reactivate(x) ;
    }
  }
}


void RemapState :: vecScatterLoop(char* locData, int remOffset,
                                  const int d, const int r) {

  if(d == rank) {
    if(apg->locIndex() == 0) {
      // This is the lead process of the active process group.

      for(remProcGrp->enumerate() ; remProcGrp->test() ; remProcGrp->next())
	exchange->putArray(typeLen, r, exts,
			   locData, locSteps,
			   remProcGrp->procId(), NULL, remOffset, remSteps) ;
    }
  }
  else {

    Range*  x      = locRngs [d] ;
    Stride* locStr = locStrs [d] ;

    Range*  y      = remRngs [d] ;
    Stride* remStr = remStrs [d] ;

    Array1<int>* subscript = subscripts [d] ;

    if(subscript) {

      // Dimension has a vector subscript

      for(x->where() ; (*x) ; (*x)++) {
	char* locDataElem = locData + locStr->offset(x) ;

        Embed ySub((*subscript)(*x), *y) ;

	if(y->distributed())
	  remProcGrp->deactivate(&ySub) ;

	int remOffsetElem = remOffset + remStr->offset(&ySub) ;

	vecScatterLoop(locDataElem, remOffsetElem, d + 1, r) ;

	if(y->distributed())
	  remProcGrp->reactivate(&ySub) ;
      } x->erewh() ;
    }
    else {

      // Dimension is unsubscripted (essentially identical to `remap').

      if(x->distributed())
	apg->deactivate(x) ;

      for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
	char* locDataElem = locData + locStr->offset(x) ;
	const int locStep = locStr->step(x) ;

	locSteps [r] = locStep ;

	Range ySub(x->blockCount(), x->blockBase(), 1, *y) ;

	for(ySub.allBlocksEnum() ; ySub.allBlocksTest() ;
				   ySub.allBlocksNext()) {
	  const int iCount = ySub.blockCount() ;

	  const int remOffsetElem = remOffset + remStr->offset(&ySub) ;
	  const int remStep       = remStr->step(&ySub) ;

	  if(y->distributed())
	    remProcGrp->deactivate(&ySub) ;

	  exts     [r] = iCount ;
	  remSteps [r] = remStep ;

	  vecScatterLoop(locDataElem, remOffsetElem, d + 1, r + 1) ;

	  if(y->distributed())
	    remProcGrp->reactivate(&ySub) ;

	  locDataElem += iCount * locStep ;
	}
      }

      if(x->distributed())
	apg->reactivate(x) ;
    }
  }
}


void RemapState :: shiftGetSubrange(char* locData, int locStep,
                          int remOffset, Range* remRng, Stride* remStr,
                          const int d,
                          const int dim, const int shift,
                          const int mode) {

  for(remRng->allBlocksEnum() ; remRng->allBlocksTest() ;
			        remRng->allBlocksNext()) {
    const int iCount = remRng->blockCount() ;

    const int remOffsetElem = remOffset + remStr->offset(remRng) ;
    const int remStep       = remStr->step(remRng) ;

    if(remRng->distributed())
      remProcGrp->deactivate(remRng) ;

    exts     [d] = iCount ;
    remSteps [d] = remStep ;

    shiftLoop(locData, remOffsetElem, d + 1, dim, shift, mode) ;

    if(remRng->distributed())
      remProcGrp->reactivate(remRng) ;

    locData += iCount * locStep ;
  }
}


void RemapState :: shiftLoop(char* locData, int remOffset, const int d,
                             const int dim, const int shift,
                             const int mode) {

  Range*  x      = locRngs [d] ;
  Stride* locStr = locStrs [d] ;

  Range*  y      = remRngs [d] ;
  Stride* remStr = remStrs [d] ;

  if(x->distributed())
    apg->deactivate(x) ;

  for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
    char* locDataElem = locData + locStr->offset(x) ;
    const int locStep = locStr->step(x) ;

    locSteps [d] = locStep ;

    if(d < dim - 1) {
      Range ySub(x->blockCount(), x->blockBase(), 1, *y) ;

      shiftGetSubrange(locDataElem, locStep,
		       remOffset, &ySub, remStr,
		       d, dim, shift, mode) ;
    }
    else {
      const int extent = x->size() ;

      switch(mode) {
	case CYCL : {
	  const int shiftMod = shift >= 0 ? (shift % extent) :
					    (shift % extent + extent) ;

	  const int yRangeLo = x->blockBase() + shiftMod ;
	  const int yRangeHi = yRangeLo + x->blockCount() - 1 ;

	  if(yRangeLo < extent) {
	    const int ySubRangeHi = min(extent - 1, yRangeHi) ;
	    const int ySubExtent  = ySubRangeHi - yRangeLo + 1 ;

	    Range ySub(ySubExtent, yRangeLo, 1, *y) ;

	    remapGetSubrange(locDataElem, locStep,
			     remOffset, &ySub, remStr, d) ;

	    locDataElem += ySubExtent * locStep ;
	  }

	  if(yRangeHi >= extent) {
	    const int ySubRangeLo = max(extent, yRangeLo) ;
	    const int ySubExtent  = yRangeHi - ySubRangeLo + 1 ;

	    Range ySub(ySubExtent, ySubRangeLo - extent, 1, *y) ;

	    remapGetSubrange(locDataElem, locStep,
			     remOffset, &ySub, remStr, d) ;
	  }

	  break ;
        }
        case EDGE : {
	  int yRangeLo = x->blockBase() + shift ;
	  int yRangeHi = yRangeLo + x->blockCount() - 1 ;

	  if(yRangeLo < 0) {
	    locDataElem -= locStep * yRangeLo ;
	    yRangeLo     = 0 ;
	  }

          if(yRangeHi >= extent)
            yRangeHi     = extent - 1 ;

	  if(yRangeHi >= yRangeLo) {
	    const int ySubExtent  = yRangeHi - yRangeLo + 1 ;

	    Range ySub(ySubExtent, yRangeLo, 1, *y) ;

	    remapGetSubrange(locDataElem, locStep,
			     remOffset, &ySub, remStr, d) ;
	  }

          break ;
	}
      }
    }
  }

  if(x->distributed())
    apg->reactivate(x) ;
}


void RemapState :: skewGetSubrange(char* locData, int locStep,
                          int remOffset, Range* remRng, Stride* remStr) {

  for(remRng->allBlocksEnum() ; remRng->allBlocksTest() ;
			        remRng->allBlocksNext()) {
    const int iCount = remRng->blockCount() ;

    const int remOffsetElem = remOffset + remStr->offset(remRng) ;
    const int remStep       = remStr->step(remRng) ;

    if(remRng->distributed())
      remProcGrp->deactivate(remRng) ;

    exts     [0] = iCount ;
    remSteps [0] = remStep ;

    exchange->getArray(typeLen, 1, exts, locData, locSteps,
                       remProcGrp->procId(), NULL, remOffsetElem, remSteps) ;

    if(remRng->distributed())
      remProcGrp->reactivate(remRng) ;

    locData += iCount * locStep ;
  }
}


void RemapState :: skewLoop(char* locData, int remOffset,
                            const int d, const int r,
                            const int dim,
                            char* shfData, Stride** shfStrs, 
                            const int mode) {

  if(d == rank) {
    // Shift 1d array in dimension `dim'.

    const int shift = *((int *) shfData) ;

    Range*  x      = locRngs [dim - 1] ;
    Stride* locStr = locStrs [dim - 1] ;

    Range*  y      = remRngs [dim - 1] ;
    Stride* remStr = remStrs [dim - 1] ;


    if(x->distributed())
      apg->deactivate(x) ;

    for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
      char* locDataElem = locData + locStr->offset(x) ;
      const int locStep = locStr->step(x) ;

      locSteps [0] = locStep ;

      const int extent = x->size() ;

      switch(mode) {
	case CYCL : {
	  const int shiftMod = shift >= 0 ? (shift % extent) :
					    (shift % extent + extent) ;

	  const int yRangeLo = x->blockBase() + shiftMod ;
	  const int yRangeHi = yRangeLo + x->blockCount() - 1 ;

	  if(yRangeLo < extent) {
	    const int ySubRangeHi = min(extent - 1, yRangeHi) ;
	    const int ySubExtent  = ySubRangeHi - yRangeLo + 1 ;

	    Range ySub(ySubExtent, yRangeLo, 1, *y) ;

	    skewGetSubrange(locDataElem, locStep, remOffset, &ySub, remStr) ;

	    locDataElem += ySubExtent * locStep ;
	  }

	  if(yRangeHi >= extent) {
	    const int ySubRangeLo = max(extent, yRangeLo) ;
	    const int ySubExtent  = yRangeHi - ySubRangeLo + 1 ;

	    Range ySub(ySubExtent, ySubRangeLo - extent, 1, *y) ;

	    skewGetSubrange(locDataElem, locStep, remOffset, &ySub, remStr) ;
	  }

	  break ;
	}
	case EDGE : {
	  int yRangeLo = x->blockBase() + shift ;
	  int yRangeHi = yRangeLo + x->blockCount() - 1 ;

	  if(yRangeLo < 0) {
	    locDataElem -= locStep * yRangeLo ;
	    yRangeLo     = 0 ;
	  }

	  if(yRangeHi >= extent)
	    yRangeHi     = extent - 1 ;

	  if(yRangeHi >= yRangeLo) {
	    const int ySubExtent  = yRangeHi - yRangeLo + 1 ;

	    Range ySub(ySubExtent, yRangeLo, 1, *y) ;

	    skewGetSubrange(locDataElem, locStep, remOffset, &ySub, remStr) ;
	  }

	  break ;
	}
      }
    }

    if(x->distributed())
      apg->reactivate(x) ;
  }
  else
    if(d != dim - 1) {
      // Loop over dimension `d'.

      Range*  x      = locRngs [d] ;
      Stride* locStr = locStrs [d] ;
      Stride* shfStr = shfStrs [r] ;

      Range*  y      = remRngs [d] ;
      Stride* remStr = remStrs [d] ;

      for(x->where() ; (*x) ; (*x)++) {
	char* locDataElem = locData + locStr->offset(x) ;
	char* shfDataElem = shfData + shfStr->offset(x) ;

	Embed ySub((*x)(), *y) ;

	if(y->distributed())
	  remProcGrp->deactivate(&ySub) ;

	int remOffsetElem = remOffset + remStr->offset(&ySub) ;

	skewLoop(locDataElem, remOffsetElem, d + 1, r + 1, dim,
                 shfDataElem, shfStrs, mode) ;

	if(y->distributed())
	  remProcGrp->reactivate(&ySub) ;
      } x->erewh() ;
    }
    else
      // d == dim - 1: Loop over any remaining dimensions.

      skewLoop(locData, remOffset, d + 1, r, dim, shfData, shfStrs, mode) ;
}




CopyState :: CopyState(const int _typeLen, const int _rank,
		       Range** _rngs, Stride** _dstStrs, Stride** _srcStrs) :
	      typeLen(_typeLen), rank(_rank),
	      rngs(_rngs), dstStrs(_dstStrs), srcStrs(_srcStrs) {}

void CopyState :: copyLoop(char* dstData, char* srcData, const int d) {

  if(d == rank)
    memcpy(dstData, srcData, typeLen) ;
  else {
    Range*  x      = rngs [d] ;

    Stride* dstStr = dstStrs [d] ;
    Stride* srcStr = srcStrs [d] ;

    for(x->locBlocksEnum() ; x->locBlocksTest() ; x->locBlocksNext()) {
      char* dstDataElem = dstData + dstStr->offset(x) ;
      char* srcDataElem = srcData + srcStr->offset(x) ;

      const int dstStep = dstStr->step(x) ;
      const int srcStep = srcStr->step(x) ;

      for(int i = 0 ; i < x->blockCount() ; i++) {
	copyLoop(dstDataElem, srcDataElem, d + 1) ;

	dstDataElem += dstStep ;
	srcDataElem += srcStep ;
      }
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `copy', `remap', `vecGather', `vecScatter', `shift', `skew'         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


void copy(Frame* dst, Frame* src) {
  CopyState state(dst->typeLen, dst->rank,
		  dst->rngs, dst->strs, src->strs) ;

  if(dst->prc()->on()) {
    state.copyLoop(dst->data, src->data, 0) ;
  } dst->prc()->no() ;
}


void remap(Frame* dst, Frame* src) {
  ProcTree spanTree ;
  apg->spanTree(&spanTree) ;

  DataExchange exchange(&spanTree, src->data) ;

  RemapState state(dst->typeLen, dst->rank,
		   dst->rngs, dst->strs, src->rngs, src->strs,
		   src->prc(), &exchange, NULL) ;

  if(dst->prc()->on()) {
    state.remapLoop(dst->data, 0, 0) ;
  } dst->prc()->no() ;
}


void vecGather(Frame* dst, Frame* src, Array1<int>** subscripts) {
  ProcTree spanTree ;
  apg->spanTree(&spanTree) ;

  DataExchange exchange(&spanTree, src->data) ;

  RemapState state(dst->typeLen, dst->rank,
		   dst->rngs, dst->strs, src->rngs, src->strs,
		   src->prc(), &exchange, subscripts) ;

  if(dst->prc()->on()) {
    state.vecGatherLoop(dst->data, 0, 0, 0) ;
  } dst->prc()->no() ;
}


void vecScatter(Frame* src, Frame* dst, Array1<int>** subscripts) {
  ProcTree spanTree ;
  apg->spanTree(&spanTree) ;

  DataExchange exchange(&spanTree, dst->data) ;

  RemapState state(src->typeLen, src->rank,
		   src->rngs, src->strs, dst->rngs, dst->strs,
		   dst->prc(), &exchange, subscripts) ;

  if(src->prc()->on()) {
    state.vecScatterLoop(src->data, 0, 0, 0) ;
  } src->prc()->no() ;
}


void shift(Frame* dst, Frame* src,
           const int shift, const int dim, const int mode) {

  ProcTree spanTree ;
  apg->spanTree(&spanTree) ;

  DataExchange exchange(&spanTree, src->data) ;

  RemapState state(dst->typeLen, dst->rank,
		   dst->rngs, dst->strs, src->rngs, src->strs,
		   src->prc(), &exchange, NULL) ;

  if(dst->prc()->on()) {
    state.shiftLoop(dst->data, 0, 0, dim, shift, mode) ;
  } dst->prc()->no() ;
}


void skew(Frame* dst, Frame* src, Frame* shift,
          const int dim, const int mode) {

  ProcTree spanTree ;
  apg->spanTree(&spanTree) ;

  DataExchange exchange(&spanTree, src->data) ;

  RemapState state(dst->typeLen, dst->rank,
		   dst->rngs, dst->strs, src->rngs, src->strs,
		   src->prc(), &exchange, NULL) ;

  if(dst->prc()->on()) {
    state.skewLoop(dst->data, 0, 0, 0, dim, shift->data, shift->strs, mode) ;
  } dst->prc()->no() ;
}


static void gatherLoop(char* locData, char** subsData,
		       Frame* loc, Frame* rem, Array<int>** subs,
		       DataExchange* exchange, int d) {

  if(d == loc->rnk()) {
    int** is = (int**) subsData ;

    ProcessGroup* remProcGrp = rem->prc();

    int r = 0, offset = 0 ;

    while(r < rem->rnk()) {

      Range* x = rem->rng(r + 1) ;
      Embed e(*(is [r]), *x) ;

      if(x->distributed())
	remProcGrp->deactivate(&e) ;
      offset += rem->str(r + 1)->offset(&e) ;

      r++ ;
    }

    exchange->get(locData, loc->len(), remProcGrp->procId(), NULL, offset) ;

    while(r > 0) {
      r-- ;

      Range* x = rem->rng(r + 1) ;

      if(x->distributed())
	remProcGrp->reactivate(x) ;
    }
  }
  else {
    Range* x = loc->rng(d + 1) ;

    for(x->where() ; (*x) ; (*x)++) {
      char* locDataElem = locData + loc->str(d + 1)->offset(x) ;

      char** subsDataElem = new char* [rem->rnk()] ;
      for(int r = 0 ; r < rem->rnk() ; r++)
        subsDataElem [r] = subsData [r] +
                           (subs [r])->str(d + 1)->offset(x) ;

      gatherLoop(locDataElem, subsDataElem, loc, rem, subs, exchange, d + 1) ;

      delete subsDataElem ;
    } x->erewh() ;
  }
}


void gather(Frame* dst, Frame* src, Array<int>** subscripts) {
  ProcTree spanTree ;
  apg->spanTree(&spanTree) ;

  DataExchange exchange(&spanTree, src->dat()) ;

  if(dst->prc()->on()) {
    char** subsData = new char* [src->rnk()] ;
    for(int r = 0 ; r < src->rnk() ; r++)
      subsData [r] = (subscripts [r])->dat() ;

    gatherLoop(dst->dat(), subsData, dst, src, subscripts, &exchange, 0) ;

    delete subsData ;
  } dst->prc()->no() ;
}


static void reshapeLoop(char* locData, int seqIndex, int seqStride,
		        Frame* loc, Frame* rem,
		        DataExchange* exchange, int d) {

  if(d == loc->rnk()) {
    ProcessGroup* remProcGrp = rem->prc();

    int r = 0, offset = 0 ;

    while(r < rem->rnk()) {

      Range* y = rem->rng(r + 1) ;

      Embed e(seqIndex % y->size(), *y) ;

      if(y->distributed())
	remProcGrp->deactivate(&e) ;
      offset += rem->str(r + 1)->offset(&e) ;

      r++ ;

      seqIndex /= y->size() ;
    }

    exchange->get(locData, loc->len(), remProcGrp->procId(), NULL, offset) ;

    while(r > 0) {
      r-- ;

      Range* y = rem->rng(r + 1) ;

      if(y->distributed())
	remProcGrp->reactivate(y) ;
    }
  }
  else {
    Range* x = loc->rng(d + 1) ;

    for(x->where() ; (*x) ; (*x)++) {
      
      char* locDataElem = locData + loc->str(d + 1)->offset(x) ;

      reshapeLoop(locDataElem,
                  seqIndex + seqStride * (*x)(), x->size() * seqStride,
		  loc, rem, exchange, d + 1) ;
    } x->erewh() ;
  }
}


//void reshape(Frame* dst, Frame* src) {
//
//  // Simple, sensible `reshape'
//
//  ProcTree spanTree ;
//  apg.spanTree(&spanTree) ;
//  DataExchange exchange(&spanTree, src->dat()) ;
//
//  if(dst->prc()->on()) {
//    reshapeLoop(dst->dat(), 0, 1, dst, src, &exchange, 0) ;
//  } dst->prc()->no() ;
//}


static void reshapePadLoop(char* locData, int seqIndex, int seqStride,
			   Frame* loc,
			   const int remSize, Frame* pad,
			   DataExchange* exchange, int d) {

  if(d == loc->rnk()) {
    if(seqIndex >= remSize) {
      Range* y = pad->rng(1) ;
      const int padIndex = (seqIndex - remSize) % y->size() ;
 
      Embed e(padIndex, *y) ;

      // Should use a *new* ProcessGroup here?
      ProcessGroup* padProcGrp = pad->prc();
      if(y->distributed())
	padProcGrp->deactivate(&e) ;

      exchange->get(locData, loc->len(), padProcGrp->procId(),
                    NULL, pad->str(1)->offset(&e)) ;

      if(y->distributed())
	padProcGrp->reactivate(y) ;
    }
  }
  else {
    Range* x = loc->rng(d + 1) ;

    for(x->where() ; (*x) ; (*x)++) {
      char* locDataElem = locData + loc->str(d + 1)->offset(x) ;

      reshapePadLoop(locDataElem,
		     seqIndex + seqStride * (*x)(), x->size() * seqStride,
		     loc, remSize, pad, exchange, d + 1) ;
    } x->erewh() ;
  }
}


void reshape(Frame* dst, Frame* src, Frame* pad) {

  // Silly, padded `reshape', to support Fortran.

  ProcTree spanTree ;
  apg->spanTree(&spanTree) ;

  {
    DataExchange exchange(&spanTree, src->dat()) ;

    if(dst->prc()->on()) {
      reshapeLoop(dst->dat(), 0, 1, dst, src, &exchange, 0) ;
    } dst->prc()->no() ;
  }

  if(pad) {
    DataExchange exchange(&spanTree, pad->dat()) ;

    int srcSize = 1 ;
    for(int d = 1 ; d <= src->rnk() ; d++)
      srcSize *= src->rng(d)->size() ;

    if(dst->prc()->on()) {
      reshapePadLoop(dst->dat(), 0, 1, dst, srcSize, pad, &exchange, 0) ;
    } dst->prc()->no() ;
  }
}


// -------------------------------------------------------------
// Things to do
//
//   Unify the implementations of the `remap', `vecGather',
//     `vecScatter' ``loops''?
//
//   Replace array get and put calls with calls to scalar versions
//     when `r' == 0.
//
//   Possibly eliminate the `RemapState' class; revert to passing
//     all information through argument list rather than using global
//     variables.  Especially, this should be done for `skew'.

